@@ -150,6 +150,11 @@ class core {
     };
     stats                    m_stats;
     friend class new_lemma;
+
+    unsigned m_nlsat_delay { 50 };
+    unsigned m_nlsat_fails { 0 };
+    bool should_run_bounded_nlsat();
+    lbool bounded_nlsat();
 public:
     var_eqs<emonics>         m_evars;
     lp::lar_solver&          m_lar_solver;
