@@ -24,6 +24,7 @@ Revision History:
 #include "ast/rewriter/th_rewriter.h"
 #include "ast/array_decl_plugin.h"
 #include "ast/bv_decl_plugin.h"
+#include "ast/recfun_decl_plugin.h"
 #include "ast/well_sorted.h"
 #include "ast/used_symbols.h"
 #include "ast/for_each_expr.h"
@@ -49,7 +50,7 @@ model::model(ast_manager & m):
 model::~model() {
     for (auto & kv : m_usort2universe) {
         m.dec_ref(kv.m_key);
-        m.dec_array_ref(kv.m_value->size(), kv.m_value->c_ptr());
+        m.dec_array_ref(kv.m_value->size(), kv.m_value->data());
         dealloc(kv.m_value);
     }
 }
@@ -73,7 +74,7 @@ void model::copy_func_interps(model const & source) {
 
 void model::copy_usort_interps(model const & source) {
     for (auto const& kv : source.m_usort2universe) 
-        register_usort(kv.m_key, kv.m_value->size(), kv.m_value->c_ptr());
+        register_usort(kv.m_key, kv.m_value->size(), kv.m_value->data());
 }
 
 model * model::copy() const {
@@ -125,6 +126,10 @@ expr * model::get_fresh_value(sort * s) {
     return get_factory(s)->get_fresh_value(s);
 }
 
+void model::register_value(expr* e) {
+    get_factory(e->get_sort())->register_value(e);
+}
+
 bool model::get_some_values(sort * s, expr_ref& v1, expr_ref& v2) {
     return get_factory(s)->get_some_values(s, v1, v2);
 }
@@ -157,7 +162,7 @@ void model::register_usort(sort * s, unsigned usize, expr * const * universe) {
         u->append(usize, universe);
     }
     else {
-        m.dec_array_ref(u->size(), u->c_ptr());
+        m.dec_array_ref(u->size(), u->data());
         u->reset();
         u->append(usize, universe);
     }
@@ -186,7 +191,7 @@ model * model::translate(ast_translation & translator) const {
         }
         res->register_usort(translator(kv.m_key),
                             new_universe.size(),
-                            new_universe.c_ptr());
+                            new_universe.data());
     }
 
     return res;
@@ -217,8 +222,6 @@ struct model::top_sort : public ::top_sort<func_decl> {
         m_occur_count.find(f, count);
         return count;
     }
-
-    ~top_sort() override {}
 };
 
 void model::compress() {
@@ -467,7 +470,7 @@ expr_ref model::cleanup_expr(top_sort& ts, expr* e, unsigned current_partition)
             else if (f->is_skolem() && can_inline_def(ts, f) && (fi = get_func_interp(f)) && 
                      fi->get_interp() && (!ts.partition_ids().find(f, pid) || pid != current_partition)) {
                 var_subst vs(m, false);
-                new_t = vs(fi->get_interp(), args.size(), args.c_ptr());
+                new_t = vs(fi->get_interp(), args.size(), args.data());
             }
             else if (bv.is_bit2bool(t)) {
                 unsigned idx = f->get_parameter(0).get_int();
@@ -480,7 +483,7 @@ expr_ref model::cleanup_expr(top_sort& ts, expr* e, unsigned current_partition)
             }
 #endif
             else {
-                new_t = ts.m_rewrite.mk_app(f, args.size(), args.c_ptr());                
+                new_t = ts.m_rewrite.mk_app(f, args.size(), args.data());                
             }
             
             if (t != new_t.get()) trail.push_back(new_t);
@@ -580,3 +583,32 @@ void model::reset_eval_cache() {
     m_mev.reset();
 }
 
+void model::add_rec_funs() {
+    recfun::util u(m);
+    func_decl_ref_vector recfuns = u.get_rec_funs();
+    for (func_decl* f : recfuns) {
+        auto& def = u.get_def(f);
+        expr* rhs = def.get_rhs();
+        if (!rhs) 
+            continue;
+        if (has_interpretation(f))
+            continue;
+        if (f->get_arity() == 0) {
+            register_decl(f, rhs);
+            continue;
+        }
+                
+        func_interp* fi = alloc(func_interp, m, f->get_arity());
+        // reverse argument order so that variable 0 starts at the beginning.
+        expr_ref_vector subst(m);
+        for (unsigned i = 0; i < f->get_arity(); ++i) {
+            subst.push_back(m.mk_var(i, f->get_domain(i)));
+        }
+        var_subst sub(m, true);
+        expr_ref bodyr = sub(rhs, subst);
+        
+        fi->set_else(bodyr);
+        register_decl(f, fi);
+    }
+    TRACE("model", tout << *this << "\n";);
+}
