@@ -304,7 +304,7 @@ namespace arith {
         return lp::EQ;
     }
 
-    void solver::mk_eq_axiom(bool is_eq, euf::th_eq const& e) {
+    void solver::new_eq_eh(euf::th_eq const& e) {
         theory_var v1 = e.v1();
         theory_var v2 = e.v2();
         if (is_bool(v1))
@@ -316,23 +316,35 @@ namespace arith {
         if (e1->get_id() > e2->get_id())
             std::swap(e1, e2);
             
-        if (is_eq && m.are_equal(e1, e2))
+        if (m.are_equal(e1, e2))
             return;
-        if (!is_eq && m.are_distinct(e1, e2))
-            return;       
 
-        if (is_eq) {       
-            ++m_stats.m_assert_eq;
-            m_new_eq = true;
-            euf::enode* n1 = var2enode(v1);
-            euf::enode* n2 = var2enode(v2);
-            lpvar w1 = register_theory_var_in_lar_solver(v1);
-            lpvar w2 = register_theory_var_in_lar_solver(v2);
-            auto cs = lp().add_equality(w1, w2);            
-            add_eq_constraint(cs.first, n1, n2);
-            add_eq_constraint(cs.second, n1, n2);
+        ++m_stats.m_assert_eq;
+        m_new_eq = true;
+        euf::enode* n1 = var2enode(v1);
+        euf::enode* n2 = var2enode(v2);
+        lpvar w1 = register_theory_var_in_lar_solver(v1);
+        lpvar w2 = register_theory_var_in_lar_solver(v2);
+        auto cs = lp().add_equality(w1, w2);            
+        add_eq_constraint(cs.first, n1, n2);
+        add_eq_constraint(cs.second, n1, n2);
+    }
+
+    void solver::new_diseq_eh(euf::th_eq const& e) {
+        m_delayed_eqs.push_back(std::make_pair(e, false));
+        ctx.push(push_back_vector<svector<std::pair<euf::th_eq, bool>>>(m_delayed_eqs));
+    }
+
+    void solver::mk_diseq_axiom(euf::th_eq const& e) {
+        if (is_bool(e.v1()))
             return;
-        }
+        force_push();
+        expr* e1 = var2expr(e.v1());
+        expr* e2 = var2expr(e.v2());
+        if (e1->get_id() > e2->get_id())
+            std::swap(e1, e2);
+        if (m.are_distinct(e1, e2))
+            return;       
         literal le, ge;
         if (a.is_numeral(e1))
             std::swap(e1, e2);
@@ -347,9 +359,7 @@ namespace arith {
             expr_ref zero(a.mk_numeral(rational(0), a.is_int(e1)), m);
             rewrite(diff);
             if (a.is_numeral(diff)) {
-                if (is_eq && a.is_zero(diff))
-                    return;
-                if (!is_eq && !a.is_zero(diff))
+                if (!a.is_zero(diff))
                     return;
                 if (a.is_zero(diff))
                     add_unit(eq);
@@ -405,15 +415,7 @@ namespace arith {
             expr* n = m_idiv_terms[i];
             expr* p = nullptr, * q = nullptr;
             VERIFY(a.is_idiv(n, p, q));
-            euf::enode* np = ctx.get_enode(p);
-            euf::enode* nn = ctx.get_enode(n);
-            if (!np || !np->is_attached_to(get_id()))
-                continue;
-            if (!nn || !nn->is_attached_to(get_id()))
-                continue;
-            theory_var v1 = mk_evar(p);
-            if (!is_registered_var(v1))
-                continue;
+            theory_var v1 = internalize_def(p);
             lp::impq r1 = get_ivalue(v1);
             rational r2;
 
@@ -432,9 +434,7 @@ namespace arith {
                     TRACE("arith", tout << "unbounded " << expr_ref(n, m) << "\n";);
                     continue;
                 }
-                theory_var v = mk_evar(n);
-                if (!is_registered_var(v))
-                    continue;
+                theory_var v = internalize_def(n);
                 lp::impq val_v = get_ivalue(v);
                 if (val_v.y.is_zero() && val_v.x == div(r1.x, r2)) continue;
 
@@ -472,4 +472,27 @@ namespace arith {
         return all_divs_valid;
     }
 
+    void solver::fixed_var_eh(theory_var v, lp::constraint_index ci1, lp::constraint_index ci2, rational const& bound) {
+        theory_var w = euf::null_theory_var;
+        enode* x = var2enode(v);
+        if (bound.is_zero()) 
+            w = lp().local_to_external(get_zero(a.is_int(x->get_expr())));
+        else if (bound.is_one())
+            w = lp().local_to_external(get_one(a.is_int(x->get_expr())));
+        else if (!m_value2var.find(bound, w))
+            return;
+        enode* y = var2enode(w);
+        if (x->get_sort() != y->get_sort())
+            return;
+        if (x->get_root() == y->get_root())
+            return;
+        reset_evidence();
+        set_evidence(ci1, m_core, m_eqs);
+        set_evidence(ci2, m_core, m_eqs);
+        ++m_stats.m_fixed_eqs;
+        auto* jst = euf::th_explain::propagate(*this, m_core, m_eqs, x, y);
+        ctx.propagate(x, y, jst->to_index());
+    }
+
+
 }
