@@ -47,7 +47,7 @@ namespace arith {
             get_one(false);
             get_zero(true);
             get_zero(false);
-            ctx.push(value_trail<euf::solver, bool>(m_internalize_initialized));
+            ctx.push(value_trail<bool>(m_internalize_initialized));
             m_internalize_initialized = true;
         }
     }
@@ -89,7 +89,7 @@ namespace arith {
     }
 
     void solver::found_unsupported(expr* n) {
-        ctx.push(value_trail<euf::solver, expr*>(m_not_handled));
+        ctx.push(value_trail<expr*>(m_not_handled));
         TRACE("arith", tout << "unsupported " << mk_pp(n, m) << "\n";);
         m_not_handled = n;
     }
@@ -125,7 +125,7 @@ namespace arith {
         if (var != UINT_MAX) {
             return var;
         }
-        ctx.push(value_trail<euf::solver, lpvar>(var));
+        ctx.push(value_trail<lpvar>(var));
         app_ref cnst(a.mk_numeral(rational(c), is_int), m);
         mk_enode(cnst);
         theory_var v = mk_evar(cnst);
@@ -164,7 +164,7 @@ namespace arith {
         unsigned index = 0;
         while (index < terms.size()) {
             SASSERT(index >= vars.size());
-            expr* n = terms[index].get();
+            expr* n = terms.get(index);
             st.to_ensure_enode().push_back(n);
             if (a.is_add(n)) {
                 for (expr* arg : *to_app(n)) {
@@ -382,7 +382,7 @@ namespace arith {
         updt_unassigned_bounds(v, +1);
         m_bounds_trail.push_back(v);
         m_bool_var2bound.insert(bv, b);
-        TRACE("arith_verbose", tout << "Internalized " << bv << ": " << mk_pp(atom, m) << "\n";);
+        TRACE("arith_verbose", tout << "Internalized " << lit << ": " << mk_pp(atom, m) << " " << *b << "\n";);
         m_new_bounds.push_back(b);
         //add_use_lists(b);
         return true;
@@ -391,7 +391,7 @@ namespace arith {
 
     bool solver::internalize_term(expr* term) {
         if (!has_var(term))
-            internalize_def(term);
+            register_theory_var_in_lar_solver(internalize_def(term));
         return true;
     }
 
@@ -447,7 +447,7 @@ namespace arith {
                 vars.push_back(register_theory_var_in_lar_solver(w));
             ensure_nla();
             m_solver->register_existing_terms();
-            m_nla->add_monic(register_theory_var_in_lar_solver(v), vars.size(), vars.c_ptr());
+            m_nla->add_monic(register_theory_var_in_lar_solver(v), vars.size(), vars.data());
         }
         return v;
     }
@@ -470,7 +470,7 @@ namespace arith {
             TRACE("arith", tout << "v" << v << " := " << mk_pp(t, m) << "\n" << vars << "\n";);
             m_solver->register_existing_terms();
             ensure_nla();
-            m_nla->add_monic(register_theory_var_in_lar_solver(v), vars.size(), vars.c_ptr());
+            m_nla->add_monic(register_theory_var_in_lar_solver(v), vars.size(), vars.data());
         }
         return v;
     }
@@ -496,6 +496,7 @@ namespace arith {
                     vi = lp().add_var(v, a.is_int(term));
                     add_def_constraint_and_equality(vi, lp::GE, st.offset());
                     add_def_constraint_and_equality(vi, lp::LE, st.offset());
+                    register_fixed_var(v, st.offset());
                     return v;
                 }
                 if (!st.offset().is_zero()) {
@@ -571,7 +572,7 @@ namespace arith {
             if (reflect(e))
                 for (expr* arg : *to_app(e))
                     args.push_back(e_internalize(arg));
-            n = ctx.mk_enode(e, args.size(), args.c_ptr());
+            n = ctx.mk_enode(e, args.size(), args.data());
             ctx.attach_node(n);
         }
         return n;
@@ -582,11 +583,10 @@ namespace arith {
         if (e->is_attached_to(get_id()))
             return e->get_th_var(get_id());
         theory_var v = mk_var(e);
-        TRACE("arith", tout << "fresh var: v" << v << " " << mk_pp(n, m) << "\n";);
+        TRACE("arith_verbose", tout << "v" << v << " " << mk_pp(n, m) << "\n";);
         SASSERT(m_bounds.size() <= static_cast<unsigned>(v) || m_bounds[v].empty());
         reserve_bounds(v);
         ctx.attach_th_var(e, this, v);
-        TRACE("arith", tout << mk_pp(n, m) << " " << v << "\n";);
         SASSERT(euf::null_theory_var != v);
         return v;
     }
@@ -674,5 +674,24 @@ namespace arith {
         return false;
     }
 
+    struct solver::undo_value : public trail {
+        solver& s;
+        undo_value(solver& s):s(s) {}
+        void undo() override {
+            s.m_value2var.erase(s.m_fixed_values.back());
+            s.m_fixed_values.pop_back();
+        }
+    };
+
+
+    void solver::register_fixed_var(theory_var v, rational const& value) {
+        if (m_value2var.contains(value)) 
+            return;
+        m_fixed_values.push_back(value);
+        m_value2var.insert(value, v);
+        ctx.push(undo_value(*this));
+    }
+
+
 }
 
