@@ -22,25 +22,43 @@ Author:
 
 namespace array {
 
+    struct solver::reset_new : trail {
+        solver& s;
+        unsigned m_idx;
+        reset_new(solver& s, unsigned idx) : s(s), m_idx(idx) {}
+        void undo() override {
+            s.m_axiom_trail[m_idx].set_new();
+        }
+    };
+
     void solver::push_axiom(axiom_record const& r) { 
         unsigned idx = m_axiom_trail.size();
         m_axiom_trail.push_back(r); 
+        TRACE("array", display(tout, r) << " " << m_axioms.contains(idx) << "\n";);
         if (m_axioms.contains(idx))
             m_axiom_trail.pop_back();
-        else
-            ctx.push(push_back_vector<euf::solver, svector<axiom_record>>(m_axiom_trail));
+        else {
+            m_axioms.insert(idx);
+            ctx.push(push_back_vector<svector<axiom_record>>(m_axiom_trail));
+            ctx.push(insert_map<axiom_table_t, unsigned>(m_axioms, idx));
+        }
     }
 
-    bool solver::propagate_axiom(unsigned idx) {
-        if (m_axioms.contains(idx))
+    bool solver::propagate_axiom(unsigned idx) {        
+        if (is_applied(idx))
             return false;
-        m_axioms.insert(idx);
-        ctx.push(insert_map<euf::solver, axiom_table_t, unsigned>(m_axioms, idx));
-        return assert_axiom(idx);
+        bool st = assert_axiom(idx);
+        if (!is_delayed(idx)) {
+            ctx.push(reset_new(*this, idx));
+            set_applied(idx);
+        }
+        return st;
     }
 
     bool solver::assert_axiom(unsigned idx) {
         axiom_record& r = m_axiom_trail[idx];
+        if (!is_relevant(r))
+            return false;
         switch (r.m_kind) {
         case axiom_record::kind_t::is_store:
             return assert_store_axiom(to_app(r.n->get_expr()));
@@ -62,8 +80,7 @@ namespace array {
     bool solver::assert_default(axiom_record& r) {
         expr* child = r.n->get_expr();
         SASSERT(can_beta_reduce(r.n));
-        if (!ctx.is_relevant(child))
-            return false;
+            
         TRACE("array", tout << "default-axiom: " << mk_bounded_pp(child, m, 2) << "\n";);
         if (a.is_const(child))
             return assert_default_const_axiom(to_app(child));
@@ -75,34 +92,42 @@ namespace array {
             return false;                
     }
 
-    struct solver::set_delay_bit : trail<euf::solver> {
-        solver& s;
-        unsigned m_idx;
-        set_delay_bit(solver& s, unsigned idx) : s(s), m_idx(idx) {}
-        void undo(euf::solver& euf) override {
-            s.m_axiom_trail[m_idx].m_delayed = false;
+
+    bool solver::is_relevant(axiom_record const& r) const {
+        return true;
+#if 0
+        // relevancy propagation is currently incomplete on terms
+
+        expr* child = r.n->get_expr();
+        switch (r.m_kind) {
+        case axiom_record::kind_t::is_select: {
+            app* select = r.select->get_app();
+            for (unsigned i = 1; i < select->get_num_args(); ++i)
+                if (!ctx.is_relevant(select->get_arg(i)))
+                    return false;
+            return ctx.is_relevant(child);            
         }
-    };
+        case axiom_record::kind_t::is_default:
+            return ctx.is_relevant(child);            
+        default:
+            return true;
+        }
+#endif
+    }
 
     bool solver::assert_select(unsigned idx, axiom_record& r) {
         expr* child = r.n->get_expr();
         app* select = r.select->get_app();
         SASSERT(a.is_select(select));
         SASSERT(can_beta_reduce(r.n));
-        if (!ctx.is_relevant(child))
-            return false;
-        for (unsigned i = 1; i < select->get_num_args(); ++i)
-            if (!ctx.is_relevant(select->get_arg(i)))
-                return false;
-        TRACE("array", tout << "select-axiom: " << mk_bounded_pp(select, m, 2) << " " << mk_bounded_pp(child, m, 2) << "\n";);
-        if (get_config().m_array_delay_exp_axiom && r.select->get_arg(0)->get_root() != r.n->get_root() && !r.m_delayed) {
+        TRACE("array", display(tout << "select-axiom: ", r) << "\n";);
+
+        if (get_config().m_array_delay_exp_axiom && r.select->get_arg(0)->get_root() != r.n->get_root() && !r.is_delayed() && m_enable_delay) {
             IF_VERBOSE(11, verbose_stream() << "delay: " << mk_bounded_pp(child, m) << " " << mk_bounded_pp(select, m) << "\n");
-            ctx.push(set_delay_bit(*this, idx));
-            r.m_delayed = true;
+            ctx.push(reset_new(*this, idx));
+            r.set_delayed();
             return false;
         }
-        if (r.select->get_arg(0)->get_root() != r.n->get_root() && r.m_delayed)
-            return false;
         if (a.is_const(child))
             return assert_select_const_axiom(select, to_app(child));
         else if (a.is_as_array(child))
@@ -144,48 +169,82 @@ namespace array {
      * where i = (i_1, ..., i_n), j = (j_1, .., j_n), k in 1..n
      */
     bool solver::assert_select_store_axiom(app* select, app* store) {
-        ++m_stats.m_num_select_store_axiom;
         SASSERT(a.is_store(store));
         SASSERT(a.is_select(select));
         SASSERT(store->get_num_args() == 1 + select->get_num_args());
         ptr_buffer<expr> sel1_args, sel2_args;
         unsigned num_args = select->get_num_args();
+
+        if (expr2enode(select->get_arg(0))->get_root() == expr2enode(store)->get_root())
+            return false;
+        bool has_diff = false;
+        for (unsigned i = 1; i < num_args; i++) 
+            has_diff |= expr2enode(select->get_arg(i))->get_root() != expr2enode(store->get_arg(i))->get_root();
+        if (!has_diff)
+            return false;
+
+
         sel1_args.push_back(store);
         sel2_args.push_back(store->get_arg(0));
-
+        
+        
         for (unsigned i = 1; i < num_args; i++) {
             sel1_args.push_back(select->get_arg(i));
             sel2_args.push_back(select->get_arg(i));
         }
-
+   
         expr_ref sel1(a.mk_select(sel1_args), m);
         expr_ref sel2(a.mk_select(sel2_args), m);
         expr_ref sel_eq_e(m.mk_eq(sel1, sel2), m);
+
+        bool new_prop = false;
+        if (!ctx.get_egraph().find(sel1))
+            new_prop = true;
+        if (!ctx.get_egraph().find(sel2))
+            new_prop = true;
+
         euf::enode* s1 = e_internalize(sel1);
         euf::enode* s2 = e_internalize(sel2);
+        TRACE("array", 
+              tout << "select-store " << ctx.bpp(s1) << " " << ctx.bpp(s1->get_root()) << "\n";
+              tout << "select-store " << ctx.bpp(s2) << " " << ctx.bpp(s2->get_root()) << "\n";);
+
+
         if (s1->get_root() == s2->get_root())
-            return false;
-        sat::literal sel_eq = mk_literal(sel_eq_e);
-        if (s().value(sel_eq) == l_true)
-            return false;
-        
-        bool new_prop = false;
+            return new_prop;
+
+        sat::literal sel_eq = sat::null_literal;
+        auto init_sel_eq = [&]() {
+            if (sel_eq != sat::null_literal) 
+                return true;
+            sel_eq = mk_literal(sel_eq_e);
+            return s().value(sel_eq) != l_true;
+        };
+
         for (unsigned i = 1; i < num_args; i++) {
             expr* idx1 = store->get_arg(i);
             expr* idx2 = select->get_arg(i);
-            euf::enode* r1 = expr2enode(idx1)->get_root();
-            euf::enode* r2 = expr2enode(idx2)->get_root();
+            euf::enode* r1 = expr2enode(idx1);
+            euf::enode* r2 = expr2enode(idx2);
             if (r1 == r2)
                 continue;
             if (m.are_distinct(r1->get_expr(), r2->get_expr())) {
-                new_prop = true;
-                add_clause(sel_eq);
+                if (init_sel_eq() && add_clause(sel_eq))
+                    new_prop = true;
                 break;
             }
             sat::literal idx_eq = eq_internalize(idx1, idx2);
+            if (s().value(idx_eq) == l_true)
+                continue;
+            if (s().value(idx_eq) == l_undef)
+                new_prop = true;
+            if (!init_sel_eq())
+                break;
             if (add_clause(idx_eq, sel_eq))
                 new_prop = true;
         }
+        ++m_stats.m_num_select_store_axiom;
+        TRACE("array", tout << "select-stored " << new_prop << "\n";);
         return new_prop;
     }
 
@@ -194,6 +253,7 @@ namespace array {
      *    select(const(v), i) = v
      */
     bool solver::assert_select_const_axiom(app* select, app* cnst) {
+        
         ++m_stats.m_num_select_const_axiom;
         expr* val = nullptr;
         VERIFY(a.is_const(cnst, val));
@@ -214,7 +274,7 @@ namespace array {
     bool solver::assert_extensionality(expr* e1, expr* e2) {
         TRACE("array", tout << "extensionality-axiom: " << mk_bounded_pp(e1, m) << " == " << mk_bounded_pp(e2, m) << "\n";);
         ++m_stats.m_num_extensionality_axiom;
-        func_decl_ref_vector const& funcs = sort2diff(m.get_sort(e1));
+        func_decl_ref_vector const& funcs = sort2diff(e1->get_sort());
         expr_ref_vector args1(m), args2(m);
         args1.push_back(e1);
         args2.push_back(e2);
@@ -223,9 +283,6 @@ namespace array {
             args1.push_back(k);
             args2.push_back(k);
         }
-        std::cout << "e1: " << mk_pp(e1, m) << "\n";
-        std::cout << "e2: " << mk_pp(e2, m) << "\n";
-        std::cout << "funcs: " << funcs << "\n";
         expr_ref sel1(a.mk_select(args1), m);
         expr_ref sel2(a.mk_select(args2), m);
         literal lit1 = eq_internalize(e1, e2);
@@ -284,7 +341,7 @@ namespace array {
         ptr_vector<expr> sel_args(num_args, select->get_args());
         sel_args[0] = arr;
         expr_ref sel(a.mk_select(sel_args), m);
-        expr_ref val(m.mk_app(f, sel_args.size() - 1, sel_args.c_ptr() + 1), m);
+        expr_ref val(m.mk_app(f, sel_args.size() - 1, sel_args.data() + 1), m);
         euf::enode* n1 = e_internalize(sel);
         euf::enode* n2 = e_internalize(val);
         return ctx.propagate(n1, n2, array_axiom());
@@ -359,7 +416,7 @@ namespace array {
 
             for (unsigned i = 1; i + 1 < num_args; ++i) {
                 expr* arg = store->get_arg(i);
-                sort* srt = m.get_sort(arg);
+                sort* srt = arg->get_sort();
                 auto ep = mk_epsilon(srt);
                 eqs.push_back(m.mk_eq(ep.first, arg));
                 args1.push_back(m.mk_app(ep.second, arg));
@@ -385,7 +442,7 @@ namespace array {
         ++m_stats.m_num_select_lambda_axiom;
         SASSERT(is_lambda(lambda));
         SASSERT(a.is_select(select));
-        SASSERT(m.get_sort(lambda) == m.get_sort(select->get_arg(0)));
+        SASSERT(lambda->get_sort() == select->get_arg(0)->get_sort());
         ptr_vector<expr> args(select->get_num_args(), select->get_args());
         args[0] = lambda;
         expr_ref alpha(a.mk_select(args), m);
@@ -400,7 +457,7 @@ namespace array {
     bool solver::assert_congruent_axiom(expr* e1, expr* e2) {
         TRACE("array", tout << "congruence-axiom: " << mk_bounded_pp(e1, m) << " " << mk_bounded_pp(e2, m) << "\n";);
         ++m_stats.m_num_congruence_axiom;
-        sort* srt         = m.get_sort(e1);
+        sort* srt         = e1->get_sort();
         unsigned dimension = get_array_arity(srt);
         expr_ref_vector args1(m), args2(m);
         args1.push_back(e1);
@@ -415,17 +472,17 @@ namespace array {
             args1.push_back(k);
             args2.push_back(k);            
         }
-        expr * sel1 = a.mk_select(dimension+1, args1.c_ptr());
-        expr * sel2 = a.mk_select(dimension+1, args2.c_ptr());
+        expr * sel1 = a.mk_select(dimension+1, args1.data());
+        expr * sel2 = a.mk_select(dimension+1, args2.data());
         expr * eq = m.mk_eq(sel1, sel2);
-        expr_ref q(m.mk_forall(dimension, sorts.c_ptr(), names.c_ptr(), eq), m);
+        expr_ref q(m.mk_forall(dimension, sorts.data(), names.data(), eq), m);
         rewrite(q);
         return add_clause(~eq_internalize(e1, e2), mk_literal(q));
     }
 
     bool solver::has_unitary_domain(app* array_term) {
         SASSERT(a.is_array(array_term));
-        sort* s = m.get_sort(array_term);
+        sort* s = array_term->get_sort();
         unsigned dim = get_array_arity(s);
         for (unsigned i = 0; i < dim; ++i) {
             sort* d = get_array_domain(s, i);
@@ -437,7 +494,7 @@ namespace array {
 
     bool solver::has_large_domain(expr* array_term) {
         SASSERT(a.is_array(array_term));
-        sort* s = m.get_sort(array_term);
+        sort* s = array_term->get_sort();
         unsigned dim = get_array_arity(s);
         rational sz(1);
         for (unsigned i = 0; i < dim; ++i) {
@@ -458,11 +515,11 @@ namespace array {
         func_decl* diag = nullptr;
         if (!m_sort2epsilon.find(s, eps)) {
             eps = m.mk_fresh_const("epsilon", s);
-            ctx.push(ast2ast_trail<euf::solver, sort, app>(m_sort2epsilon, s, eps));
+            ctx.push(ast2ast_trail<sort, app>(m_sort2epsilon, s, eps));
         }
         if (!m_sort2diag.find(s, diag)) {
             diag = m.mk_fresh_func_decl("diag", 1, &s, s);
-            ctx.push(ast2ast_trail<euf::solver, sort, func_decl>(m_sort2diag, s, diag));
+            ctx.push(ast2ast_trail<sort, func_decl>(m_sort2diag, s, diag));
         }
         return std::make_pair(eps, diag);
     }
@@ -486,9 +543,11 @@ namespace array {
         bool change = false;
         unsigned sz = m_axiom_trail.size();
         m_delay_qhead = 0;
+        
         for (; m_delay_qhead < sz; ++m_delay_qhead) 
-            if (m_axiom_trail[m_delay_qhead].m_delayed && assert_axiom(m_delay_qhead))
-                change = true;        
+            if (m_axiom_trail[m_delay_qhead].is_delayed() && assert_axiom(m_delay_qhead))
+                change = true;  
+        flet<bool> _enable_delay(m_enable_delay, false);
         if (unit_propagate())
             change = true;
         return change;
@@ -504,7 +563,7 @@ namespace array {
             for (unsigned j = i; j-- > 0; ) {
                 theory_var v2 = roots[j];
                 expr* e2 = var2expr(v2);
-                if (m.get_sort(e1) != m.get_sort(e2))
+                if (e1->get_sort() != e2->get_sort())
                     continue;
                 if (have_different_model_values(v1, v2))
                     continue;
@@ -536,7 +595,7 @@ namespace array {
             r->mark1();
             to_unmark.push_back(r);            
         }
-        TRACE("array", tout << "collecting shared vars...\n" << unsigned_vector(roots.size(), (unsigned*)roots.c_ptr())  << "\n";);
+        TRACE("array", tout << "collecting shared vars...\n" << unsigned_vector(roots.size(), (unsigned*)roots.data())  << "\n";);
         for (auto* n : to_unmark)
             n->unmark1();
     }
