@@ -92,10 +92,11 @@ namespace euf {
             m_args.push_back(m_egraph.find(to_app(e)->get_arg(i)));
         if (root && internalize_root(to_app(e), sign, m_args))
             return false;
+        SASSERT(!get_enode(e));
         if (auto* s = expr2solver(e)) 
             s->internalize(e, m_is_redundant);        
         else 
-            attach_node(m_egraph.mk(e, m_generation, num, m_args.c_ptr()));        
+            attach_node(m_egraph.mk(e, m_generation, num, m_args.data()));        
         return true;
     }
 
@@ -108,15 +109,15 @@ namespace euf {
         if (m.is_bool(e))
             attach_lit(literal(si.add_bool_var(e), false), e);
 
-        if (!m.is_bool(e) && m.get_sort(e)->get_family_id() != null_family_id) {
+        if (!m.is_bool(e) && e->get_sort()->get_family_id() != null_family_id) {
             auto* e_ext = expr2solver(e);
-            auto* s_ext = sort2solver(m.get_sort(e));
+            auto* s_ext = sort2solver(e->get_sort());
             if (s_ext && s_ext != e_ext)
-                s_ext->apply_sort_cnstr(n, m.get_sort(e));
+                s_ext->apply_sort_cnstr(n, e->get_sort());
         }
         expr* a = nullptr, * b = nullptr;                   
-        if (m.is_eq(e, a, b) && m.get_sort(a)->get_family_id() != null_family_id) {
-            auto* s_ext = sort2solver(m.get_sort(a));
+        if (m.is_eq(e, a, b) && a->get_sort()->get_family_id() != null_family_id) {
+            auto* s_ext = sort2solver(a->get_sort());
             if (s_ext)
                 s_ext->eq_internalized(n);
         }
@@ -128,6 +129,7 @@ namespace euf {
         s().set_external(v);
         s().set_eliminated(v, false);   
 
+
         if (lit.sign()) {
             v = si.add_bool_var(e);
             s().set_external(v);
@@ -139,8 +141,7 @@ namespace euf {
         }
 
         m_bool_var2expr.reserve(v + 1, nullptr);
-        if (m_bool_var2expr[v]) {
-            SASSERT(m_egraph.find(e));
+        if (m_bool_var2expr[v] && m_egraph.find(e)) {
             SASSERT(m_egraph.find(e)->bool_var() == v);
             return lit;
         }
@@ -156,6 +157,8 @@ namespace euf {
             m_egraph.set_merge_enabled(n, false);
         if (!si.is_bool_op(e))
             track_relevancy(lit.var());
+        if (s().value(lit) != l_undef) 
+            m_egraph.set_value(n, s().value(lit));
         return lit;
     }
 
@@ -163,9 +166,9 @@ namespace euf {
         if (m.is_distinct(e)) {
             enode_vector _args(args);
             if (sign)
-                add_not_distinct_axiom(e, _args.c_ptr());
+                add_not_distinct_axiom(e, _args.data());
             else
-                add_distinct_axiom(e, _args.c_ptr());
+                add_distinct_axiom(e, _args.data());
             return true;
         }
         return false;
@@ -190,12 +193,12 @@ namespace euf {
             }
             s().mk_clause(lits, st);
             if (relevancy_enabled())
-                add_root(lits.size(), lits.c_ptr());
-    }
+                add_root(lits);
+        }
         else {
             // g(f(x_i)) = x_i
             // f(x_1) = a + .... + f(x_n) = a >= 2
-            sort* srt = m.get_sort(e->get_arg(0));
+            sort* srt = e->get_arg(0)->get_sort();
             SASSERT(!m.is_bool(srt));
             sort_ref u(m.mk_fresh_sort("distinct-elems"), m);
             sort* u_ptr = u.get();
@@ -212,7 +215,7 @@ namespace euf {
                 eqs.push_back(mk_eq(fapp, a));
             }
             pb_util pb(m);
-            expr_ref at_least2(pb.mk_at_least_k(eqs.size(), eqs.c_ptr(), 2), m);
+            expr_ref at_least2(pb.mk_at_least_k(eqs.size(), eqs.data(), 2), m);
             sat::literal lit = si.internalize(at_least2, m_is_redundant);
             s().mk_clause(1, &lit, st);
             if (relevancy_enabled())
@@ -242,7 +245,7 @@ namespace euf {
         }
         else {
             // dist-f(x_1) = v_1 & ... & dist-f(x_n) = v_n
-            sort* srt = m.get_sort(e->get_arg(0));
+            sort* srt = e->get_arg(0)->get_sort();
             SASSERT(!m.is_bool(srt));
             sort_ref u(m.mk_fresh_sort("distinct-elems"), m);
             func_decl_ref f(m.mk_fresh_func_decl("dist-f", "", 1, &srt, u), m);
@@ -265,17 +268,18 @@ namespace euf {
         sat::status st = sat::status::th(m_is_redundant, m.get_basic_family_id());
         expr* c = nullptr, * th = nullptr, * el = nullptr;
         if (!m.is_bool(e) && m.is_ite(e, c, th, el)) {
-            app* a = to_app(e);
-            expr_ref eq_th = mk_eq(a, th);
+            expr_ref eq_th = mk_eq(e, th);
             sat::literal lit_th = mk_literal(eq_th);
             if (th == el) {
                 s().add_clause(1, &lit_th, st);
             }
             else {
                 sat::bool_var v = si.to_bool_var(c);
+                VERIFY(v != sat::null_bool_var);
+                VERIFY(s().is_external(v));
                 SASSERT(v != sat::null_bool_var);
-
-                expr_ref eq_el = mk_eq(a, el);
+                VERIFY(!s().was_eliminated(v));
+                expr_ref eq_el = mk_eq(e, el);
 
                 sat::literal lit_el = mk_literal(eq_el);
                 literal lits1[2] = { literal(v, true),  lit_th };
@@ -387,4 +391,25 @@ namespace euf {
             r = m.mk_eq(e1, e2);
         return r;
     }
+
+    unsigned solver::get_max_generation(expr* e) const {
+        unsigned g = 0;
+        expr_fast_mark1 mark;
+        m_todo.push_back(e);
+        while (!m_todo.empty()) {
+            e = m_todo.back();
+            m_todo.pop_back();
+            if (mark.is_marked(e))
+                continue;
+            mark.mark(e);
+            euf::enode* n = m_egraph.find(e);
+            if (n) 
+                g = std::max(g, n->generation());
+            else if (is_app(e)) 
+                for (expr* arg : *to_app(e))
+                    m_todo.push_back(arg);
+        }
+        return g;
+    }
+
 }
