@@ -35,7 +35,7 @@ namespace euf {
 
         ~user_sort() {
             for (auto kv : sort2values)
-                mdl->register_usort(kv.m_key, kv.m_value->size(), kv.m_value->c_ptr());
+                mdl->register_usort(kv.m_key, kv.m_value->size(), kv.m_value->data());
         }
 
         void add(enode* r, sort* srt) {
@@ -73,7 +73,7 @@ namespace euf {
         values2model(deps, mdl);
         for (auto* mb : m_solvers)
             mb->finalize_model(*mdl);
-        // validate_model(*mdl);
+        validate_model(*mdl);
     }
 
     bool solver::include_func_interp(func_decl* f) {
@@ -88,31 +88,43 @@ namespace euf {
     }
 
     void solver::collect_dependencies(user_sort& us, deps_t& deps) {
+        ptr_buffer<enode> fresh_values;
         for (enode* n : m_egraph.nodes()) {
             expr* e = n->get_expr();
-            sort* srt = m.get_sort(e);
+            sort* srt = e->get_sort();
             auto* mb = sort2solver(srt);
-            if (mb)
-                mb->add_dep(n, deps);
-            else
+            if (!mb) 
                 deps.insert(n, nullptr);
+            else if (!mb->add_dep(n, deps))
+                fresh_values.push_back(n);
             if (n->is_root() && m.is_uninterp(srt) && m.is_value(e))
                 us.register_value(e);
         }
 
+        // fresh values depend on all non-fresh values of the same sort
+        for (enode* n : fresh_values) {
+            n->mark1();
+            deps.insert(n, nullptr);
+        }
+        for (enode* n : fresh_values)
+            for (enode* r : m_egraph.nodes())
+                if (r->is_root() && r->get_sort() == n->get_sort() && !r->is_marked1())
+                    deps.add(n, r);
+        for (enode* n : fresh_values)
+            n->unmark1();
+        
         TRACE("euf",
               for (auto const& d : deps.deps()) 
                   if (d.m_value) {
-                      tout << mk_bounded_pp(d.m_key->get_expr(), m) << ":\n";
+                      tout << bpp(d.m_key) << ":\n";
                       for (auto* n : *d.m_value)
-                          tout << "   " << mk_bounded_pp(n->get_expr(), m) << "\n";
+                          tout << "   " << bpp(n) << "\n";
                   }
               );
     }
 
     void solver::dependencies2values(user_sort& us, deps_t& deps, model_ref& mdl) {
         for (enode* n : deps.top_sorted()) {
-            
             unsigned id = n->get_root_id();
             if (m_values.get(id, nullptr))
                 continue;
@@ -132,6 +144,16 @@ namespace euf {
                     m_values.set(id, m.mk_false());
                     continue;
                 }
+                switch (n->value()) {
+                case l_true:
+                    m_values.set(id, m.mk_true());
+                    continue;
+                case l_false:
+                    m_values.set(id, m.mk_false());
+                    continue;
+                default:
+                    break;
+                }
                 if (is_app(e) && to_app(e)->get_family_id() == m.get_basic_family_id())
                     continue;
                 sat::bool_var v = get_enode(e)->bool_var();
@@ -148,7 +170,7 @@ namespace euf {
                 }
                 continue;
             }
-            sort* srt = m.get_sort(e);
+            sort* srt = e->get_sort();
             if (m.is_uninterp(srt)) 
                 us.add(n->get_root(), srt);
             else if (auto* mbS = sort2solver(srt))
@@ -187,12 +209,18 @@ namespace euf {
                     mdl->register_decl(f, fi);
                 }
                 args.reset();                
-                for (enode* arg : enode_args(n)) 
-                    args.push_back(m_values.get(arg->get_root_id()));                
+                for (expr* arg : *a) {
+                    enode* earg = get_enode(arg); 
+                    args.push_back(m_values.get(earg->get_root_id()));                
+                }
                 DEBUG_CODE(for (expr* arg : args) VERIFY(arg););
                 SASSERT(args.size() == arity);
-                if (!fi->get_entry(args.c_ptr()))
-                    fi->insert_new_entry(args.c_ptr(), v);
+                if (!fi->get_entry(args.data()))
+                    fi->insert_new_entry(args.data(), v);
+                TRACE("euf", tout << f->get_name() << "\n";
+                      for (expr* arg : args) tout << mk_pp(arg, m) << " ";
+                      tout << "\n -> " << mk_pp(v, m) << "\n";);
+
             }
         }
     }
@@ -224,13 +252,17 @@ namespace euf {
             expr* e = n->get_expr();
             if (!m.is_bool(e))
                 continue;
-            unsigned id = n->get_root_id();
-            if (!m_values.get(id))
+            if (!is_relevant(n))
                 continue;
-            bool tt = m.is_true(m_values.get(id));
-            if (mdl.is_true(e) != tt) {
-                IF_VERBOSE(0, verbose_stream() << "Failed to evaluate " << id << " " << mk_bounded_pp(e, m) << " " << mdl(e) << " " << mk_bounded_pp(m_values.get(id), m) << "\n");
-            }
+            bool tt = l_true == s().value(n->bool_var());
+            if (tt && !mdl.is_false(e))
+                continue;
+            if (!tt && !mdl.is_true(e))
+                continue;
+            IF_VERBOSE(0, 
+                       verbose_stream() << "Failed to validate " << n->bool_var() << " " << bpp(n) << " " << mdl(e) << "\n";
+                       for (auto* arg : euf::enode_args(n))
+                           verbose_stream() << bpp(arg) << "\n" << mdl(arg->get_expr()) << "\n";);
         }
         
     }
