@@ -98,8 +98,8 @@ namespace smt {
         m_parser.add_var("cs_factor");
     }
 
-    quantifier_stat * qi_queue::set_values(quantifier * q, app * pat, unsigned generation, unsigned min_top_generation, unsigned max_top_generation, float cost) {
-        quantifier_stat * stat     = m_qm.get_stat(q);
+    q::quantifier_stat * qi_queue::set_values(quantifier * q, app * pat, unsigned generation, unsigned min_top_generation, unsigned max_top_generation, float cost) {
+        q::quantifier_stat * stat     = m_qm.get_stat(q);
         m_vals[COST]               = cost;
         m_vals[MIN_TOP_GENERATION] = static_cast<float>(min_top_generation);
         m_vals[MAX_TOP_GENERATION] = static_cast<float>(max_top_generation);
@@ -120,8 +120,8 @@ namespace smt {
     }
 
     float qi_queue::get_cost(quantifier * q, app * pat, unsigned generation, unsigned min_top_generation, unsigned max_top_generation) {
-        quantifier_stat * stat = set_values(q, pat, generation, min_top_generation, max_top_generation, 0);
-        float r = m_evaluator(m_cost_function, m_vals.size(), m_vals.c_ptr());
+        q::quantifier_stat * stat = set_values(q, pat, generation, min_top_generation, max_top_generation, 0);
+        float r = m_evaluator(m_cost_function, m_vals.size(), m_vals.data());
         stat->update_max_cost(r);
         return r;
     }
@@ -129,7 +129,7 @@ namespace smt {
     unsigned qi_queue::get_new_gen(quantifier * q, unsigned generation, float cost) {
         // max_top_generation and min_top_generation are not available for computing inc_gen
         set_values(q, nullptr, generation, 0, 0, cost);
-        float r = m_evaluator(m_new_gen_function, m_vals.size(), m_vals.c_ptr());
+        float r = m_evaluator(m_new_gen_function, m_vals.size(), m_vals.data());
         return std::max(generation + 1, static_cast<unsigned>(r));
     }
 
@@ -140,7 +140,7 @@ namespace smt {
               tout << "new instance of " << q->get_qid() << ", weight " << q->get_weight()
               << ", generation: " << generation << ", scope_level: " << m_context.get_scope_level() << ", cost: " << cost << "\n";
               for (unsigned i = 0; i < f->get_num_args(); i++) {
-                  tout << "#" << f->get_arg(i)->get_owner_id() << " d:" << f->get_arg(i)->get_owner()->get_depth() << " ";
+                  tout << "#" << f->get_arg(i)->get_expr_id() << " d:" << f->get_arg(i)->get_expr()->get_depth() << " ";
               }
               tout << "\n";);
         TRACE("new_entries_bug", tout << "[qi:insert]\n";);
@@ -206,7 +206,7 @@ namespace smt {
                 
         TRACE("qi_queue_profile", tout << q->get_qid() << ", gen: " << generation << " " << *f << " cost: " << ent.m_cost << "\n";);
 
-        quantifier_stat * stat = m_qm.get_stat(q);
+        q::quantifier_stat * stat = m_qm.get_stat(q);
 
         if (m_checker.is_sat(q->get_expr(), num_bindings, bindings)) {
             TRACE("checker", tout << "instance already satisfied\n";);
@@ -221,8 +221,10 @@ namespace smt {
 
         STRACE("instance", tout << "### " << static_cast<void*>(f) <<", " << q->get_qid()  << "\n";);
 
-        expr_ref instance(m);
-        m_subst(q, num_bindings, bindings, instance);
+        auto* ebindings = m_subst(q, num_bindings);
+        for (unsigned i = 0; i < num_bindings; ++i)
+            ebindings[i] = bindings[i]->get_expr();
+        expr_ref instance = m_subst();
 
         TRACE("qi_queue", tout << "new instance:\n" << mk_pp(instance, m) << "\n";);
         TRACE("qi_queue_instance", tout << "new instance:\n" << mk_pp(instance, m) << "\n";);
@@ -252,7 +254,7 @@ namespace smt {
             ptr_vector<expr> args;
             args.push_back(m.mk_not(q));
             args.append(to_app(s_instance)->get_num_args(), to_app(s_instance)->get_args());
-            lemma = m.mk_or(args.size(), args.c_ptr());
+            lemma = m.mk_or(args.size(), args.data());
         }
         else if (m.is_false(s_instance)) {
             lemma = m.mk_not(q);
@@ -269,10 +271,10 @@ namespace smt {
         if (m.proofs_enabled()) {
             expr_ref_vector bindings_e(m);
             for (unsigned i = 0; i < num_bindings; ++i) {
-                bindings_e.push_back(bindings[i]->get_owner());
+                bindings_e.push_back(bindings[i]->get_expr());
             }
             app * bare_lemma    = m.mk_or(m.mk_not(q), instance);
-            proof * qi_pr       = m.mk_quant_inst(bare_lemma, num_bindings, bindings_e.c_ptr());
+            proof * qi_pr       = m.mk_quant_inst(bare_lemma, num_bindings, bindings_e.data());
             proof_id            = qi_pr->get_id();
             if (bare_lemma == lemma) {
                 pr1             = qi_pr;
