@@ -132,6 +132,8 @@ namespace smt {
         if (a == b) {
             return true_literal;
         }
+        if (m.are_distinct(a, b))
+            return false_literal;
         app_ref eq(ctx.mk_eq_atom(a, b), get_manager());
         TRACE("mk_var_bug", tout << "mk_eq: " << eq->get_id() << " " << a->get_id() << " " << b->get_id() << "\n";
               tout << mk_ll_pp(a, get_manager()) << "\n" << mk_ll_pp(b, get_manager()););		
@@ -159,9 +161,9 @@ namespace smt {
     }
 
     enode* theory::ensure_enode(expr* e) {
-        if (!ctx.e_internalized(e)) {
+        if (!ctx.e_internalized(e)) 
             ctx.internalize(e, is_quantifier(e));
-        }
+        ctx.ensure_internalized(e); // make sure theory variables are attached.
         enode* n = ctx.get_enode(e);
         ctx.mark_as_relevant(n);
         return n;
