@@ -45,7 +45,7 @@ namespace smt {
     
     void theory_bv::mk_bits(theory_var v) {
         enode * n             = get_enode(v);
-        app * owner           = n->get_owner();
+        app * owner           = n->get_expr();
         unsigned bv_size      = get_bv_size(n);
         bool is_relevant      = ctx.is_relevant(n);
         literal_vector & bits = m_bits[v];
@@ -55,7 +55,7 @@ namespace smt {
         for (unsigned i = 0; i < bv_size; i++) {
             m_bits_expr.push_back(mk_bit2bool(owner, i));
         }
-        ctx.internalize(m_bits_expr.c_ptr(), bv_size, true);
+        ctx.internalize(m_bits_expr.data(), bv_size, true);
 
         for (unsigned i = 0; i < bv_size; i++) {
             bool_var b = ctx.get_bool_var(m_bits_expr[i]);
@@ -72,11 +72,12 @@ namespace smt {
 
     }
 
-    class mk_atom_trail : public trail<theory_bv> {
+    class mk_atom_trail : public trail {
+        theory_bv& th;
         bool_var m_var;
     public:
-        mk_atom_trail(bool_var v):m_var(v) {}
-        void undo(theory_bv & th) override {
+        mk_atom_trail(theory_bv& th, bool_var v):th(th), m_var(v) {}
+        void undo() override {
             theory_bv::atom * a = th.get_bv2a(m_var);
             a->~atom();
             th.erase_bv2a(m_var);
@@ -123,7 +124,7 @@ namespace smt {
             ctx.set_var_theory(bv, get_id());
             bit_atom * a     = new (get_region()) bit_atom();
             insert_bv2a(bv, a);
-            m_trail_stack.push(mk_atom_trail(bv));
+            m_trail_stack.push(mk_atom_trail(*this, bv));
             unsigned idx     = n->get_decl()->get_parameter(0).get_int();
             SASSERT(a->m_occs == 0);
             a->m_occs = new (get_region()) var_pos_occ(v_arg, idx);
@@ -179,7 +180,7 @@ namespace smt {
             return n->get_arg(idx);
         }
         else {
-            app * arg     = to_app(n->get_owner()->get_arg(idx));
+            app * arg     = to_app(n->get_expr()->get_arg(idx));
             SASSERT(ctx.e_internalized(arg));
             return ctx.get_enode(arg);
         }
@@ -212,11 +213,11 @@ namespace smt {
         get_bits(ctx.get_enode(arg), r);
     }
     
-    class add_var_pos_trail : public trail<theory_bv> {
+    class add_var_pos_trail : public trail {
         theory_bv::bit_atom * m_atom;
     public:
         add_var_pos_trail(theory_bv::bit_atom * a):m_atom(a) {}
-        void undo(theory_bv & th) override {
+        void undo() override {
             SASSERT(m_atom->m_occs);
             m_atom->m_occs = m_atom->m_occs->m_next;
         }
@@ -226,7 +227,7 @@ namespace smt {
         if (!params().m_bv_eq_axioms)
             return;
         m_prop_diseqs.push_back(bv_diseq(v1, v2, idx));
-        ctx.push_trail(push_back_vector<context, svector<bv_diseq>>(m_prop_diseqs));
+        ctx.push_trail(push_back_vector<svector<bv_diseq>>(m_prop_diseqs));
     }
 
     /**
@@ -302,7 +303,7 @@ namespace smt {
                 SASSERT(ctx.get_var_theory(l.var()) == get_id());
                 bit_atom * b = new (get_region()) bit_atom();
                 insert_bv2a(l.var(), b);
-                m_trail_stack.push(mk_atom_trail(l.var()));
+                m_trail_stack.push(mk_atom_trail(*this, l.var()));
                 SASSERT(b->m_occs == 0);
                 b->m_occs = new (get_region()) var_pos_occ(v, idx);
             }
@@ -324,7 +325,7 @@ namespace smt {
         SASSERT(get_bv_size(n) == sz);
         m_bits[v].reset();
 
-        ctx.internalize(bits.c_ptr(), sz, true);
+        ctx.internalize(bits.data(), sz, true);
 
         for (unsigned i = 0; i < sz; i++) {
             expr * bit          = bits.get(i);
@@ -417,9 +418,9 @@ namespace smt {
             }
             if (!visited)
                 return nullptr;
-            expr * fact     = ctx.mk_eq_atom(m_th.get_enode(m_var1)->get_owner(), m_th.get_enode(m_var2)->get_owner());
+            expr * fact     = ctx.mk_eq_atom(m_th.get_enode(m_var1)->get_expr(), m_th.get_enode(m_var2)->get_expr());
             ast_manager & m = ctx.get_manager();
-            return m.mk_th_lemma(get_from_theory(), fact, prs.size(), prs.c_ptr());
+            return m.mk_th_lemma(get_from_theory(), fact, prs.size(), prs.data());
         }
 
         theory_id get_from_theory() const override {
@@ -443,8 +444,8 @@ namespace smt {
             return;
         }
         ++m_stats.m_num_eq_dynamic;
-        app* o1 = get_enode(v1)->get_owner();
-        app* o2 = get_enode(v2)->get_owner();
+        app* o1 = get_enode(v1)->get_expr();
+        app* o2 = get_enode(v2)->get_expr();
         literal oeq = mk_eq(o1, o2, true);
         unsigned sz = get_bv_size(v1);
         TRACE("bv", 
@@ -470,7 +471,7 @@ namespace smt {
             eqs.push_back(~eq);
         }
         eqs.push_back(oeq);
-        ctx.mk_th_axiom(get_id(), eqs.size(), eqs.c_ptr());
+        ctx.mk_th_axiom(get_id(), eqs.size(), eqs.data());
     }
 
     void theory_bv::fixed_var_eh(theory_var v) {
@@ -478,7 +479,7 @@ namespace smt {
         VERIFY(get_fixed_value(v, val));
         enode* n = get_enode(v);
         if (ctx.watches_fixed(n)) {
-            expr_ref num(m_util.mk_numeral(val, m.get_sort(n->get_owner())), m);
+            expr_ref num(m_util.mk_numeral(val, n->get_expr()->get_sort()), m);
             literal_vector& lits = m_tmp_literals;
             lits.reset();
             for (literal b : m_bits[v]) {
@@ -602,9 +603,9 @@ namespace smt {
         SASSERT(ctx.e_internalized(n));
         SASSERT(m_util.is_bv2int(n));
         TRACE("bv2int_bug", tout << "bv2int:\n" << mk_pp(n, m) << "\n";);
-        sort * int_sort = m.get_sort(n);
+        sort * int_sort = n->get_sort();
         app * k = to_app(n->get_arg(0));
-        SASSERT(m_util.is_bv_sort(m.get_sort(k)));
+        SASSERT(m_util.is_bv_sort(k->get_sort()));
         expr_ref_vector k_bits(m);
         enode * k_enode = mk_enode(k);
         get_bits(k_enode, k_bits);
@@ -633,7 +634,7 @@ namespace smt {
             args.push_back(m.mk_ite(b, n, zero));
             num *= numeral(2);
         }
-        expr_ref sum(m_autil.mk_add(sz, args.c_ptr()), m);
+        expr_ref sum(m_autil.mk_add(sz, args.data()), m);
         th_rewriter rw(m);
         rw(sum);
         literal l(mk_eq(n, sum, false));
@@ -703,9 +704,9 @@ namespace smt {
         for (unsigned i = 0; i < sz; ++i) {
             numeral div = power(numeral(2), i);
             mod = numeral(2);
-            rhs = m_autil.mk_idiv(e, m_autil.mk_numeral(div,true));
+            rhs = (i == 0) ? e : m_autil.mk_idiv(e, m_autil.mk_numeral(div, true));
             rhs = m_autil.mk_mod(rhs, m_autil.mk_numeral(mod, true));
-            rhs = m.mk_eq(rhs, m_autil.mk_numeral(rational(1), true));
+            rhs = ctx.mk_eq_atom(rhs, m_autil.mk_int(1));
             lhs = n_bits.get(i);
             TRACE("bv", tout << mk_pp(lhs, m) << " == " << mk_pp(rhs, m) << "\n";);
             l = literal(mk_eq(lhs, rhs, false));
@@ -726,7 +727,7 @@ namespace smt {
         enode * e       = mk_enode(n);                                  \
         expr_ref_vector arg1_bits(m), bits(m);                          \
         get_arg_bits(e, 0, arg1_bits);                                  \
-        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.c_ptr(), bits);       \
+        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.data(), bits);       \
         init_bits(e, bits);                                             \
     }
 
@@ -740,7 +741,7 @@ namespace smt {
         get_arg_bits(e, 0, arg1_bits);                                                  \
         get_arg_bits(e, 1, arg2_bits);                                                  \
         SASSERT(arg1_bits.size() == arg2_bits.size());                                  \
-        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.c_ptr(), arg2_bits.c_ptr(), bits);    \
+        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.data(), arg2_bits.data(), bits);    \
         init_bits(e, bits);                                                             \
     }
 
@@ -763,7 +764,7 @@ namespace smt {
             get_arg_bits(e, i, arg_bits);                                                       \
             SASSERT(arg_bits.size() == bits.size());                                            \
             new_bits.reset();                                                                   \
-            m_bb.BLAST_OP(arg_bits.size(), arg_bits.c_ptr(), bits.c_ptr(), new_bits);           \
+            m_bb.BLAST_OP(arg_bits.size(), arg_bits.data(), bits.data(), new_bits);           \
             bits.swap(new_bits);                                                                \
         }                                                                                       \
         init_bits(e, bits);                                                                     \
@@ -780,7 +781,7 @@ namespace smt {
         get_arg_bits(e, 1, arg2_bits);                                                  
         SASSERT(arg1_bits.size() == arg2_bits.size());                                  
         expr_ref carry(m);
-        m_bb.mk_subtracter(arg1_bits.size(), arg1_bits.c_ptr(), arg2_bits.c_ptr(), bits, carry);    
+        m_bb.mk_subtracter(arg1_bits.size(), arg1_bits.data(), arg2_bits.data(), bits, carry);    
         init_bits(e, bits);                                                                
     }
 
@@ -817,7 +818,7 @@ namespace smt {
         expr_ref_vector arg1_bits(m), bits(m);                                  \
         get_arg_bits(e, 0, arg1_bits);                                          \
         unsigned param  = n->get_decl()->get_parameter(0).get_int();            \
-        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.c_ptr(), param, bits);        \
+        m_bb.BLAST_OP(arg1_bits.size(), arg1_bits.data(), param, bits);        \
         init_bits(e, bits);                                                     \
     }
     
@@ -938,7 +939,7 @@ namespace smt {
         get_arg_bits(n, 0, arg1_bits);                                                                          \
         get_arg_bits(n, 1, arg2_bits);                                                                          \
         expr_ref out(m);                                                                                        \
-        m_bb.OP(arg1_bits.size(), arg1_bits.c_ptr(), arg2_bits.c_ptr(), out);                                   \
+        m_bb.OP(arg1_bits.size(), arg1_bits.data(), arg2_bits.data(), out);                                   \
         expr_ref s_out(m);                                                                                      \
         simplify_bit(out, s_out);                                                                               \
         ctx.internalize(s_out, true);                                                                           \
@@ -947,7 +948,7 @@ namespace smt {
         ctx.set_var_theory(l.var(), get_id());                                                                  \
         le_atom * a     = new (get_region()) le_atom(l, def); /* abuse le_atom */                               \
         insert_bv2a(l.var(), a);                                                                                \
-        m_trail_stack.push(mk_atom_trail(l.var()));                                                             \
+        m_trail_stack.push(mk_atom_trail(*this, l.var()));                                                             \
         /* smul_no_overflow and umul_no_overflow are using the le_atom (THIS IS A BIG HACK)... */               \
         /* the connection between the l and def was never realized when                        */               \
         /* relevancy() is true and m_bv_lazy_le is false (the default configuration).          */               \
@@ -972,9 +973,9 @@ namespace smt {
         get_arg_bits(n, 1, arg2_bits);                                                  
         expr_ref le(m);
         if (Signed)
-            m_bb.mk_sle(arg1_bits.size(), arg1_bits.c_ptr(), arg2_bits.c_ptr(), le);
+            m_bb.mk_sle(arg1_bits.size(), arg1_bits.data(), arg2_bits.data(), le);
         else
-            m_bb.mk_ule(arg1_bits.size(), arg1_bits.c_ptr(), arg2_bits.c_ptr(), le);
+            m_bb.mk_ule(arg1_bits.size(), arg1_bits.data(), arg2_bits.data(), le);
         expr_ref s_le(m);
         simplify_bit(le, s_le);
         ctx.internalize(s_le, true);
@@ -983,7 +984,7 @@ namespace smt {
         ctx.set_var_theory(l.var(), get_id());
         le_atom * a     = new (get_region()) le_atom(l, def);
         insert_bv2a(l.var(), a);
-        m_trail_stack.push(mk_atom_trail(l.var()));
+        m_trail_stack.push(mk_atom_trail(*this, l.var()));
         if (!ctx.relevancy() || !params().m_bv_lazy_le) {
             ctx.mk_th_axiom(get_id(),  l, ~def);
             ctx.mk_th_axiom(get_id(), ~l,  def);
@@ -1086,11 +1087,11 @@ namespace smt {
         unsigned num_args = n->get_num_args();
         for (unsigned i = 0; i <= num_args; i++) {
             expr* arg = (i == num_args)?n:n->get_arg(i);
-            sort* s = m.get_sort(arg);
+            sort* s = arg->get_sort();
             if (m_util.is_bv_sort(s) && m_util.get_bv_size(arg) > params().m_bv_blast_max_size) {                
                 if (!m_approximates_large_bvs) {
                     TRACE("bv", tout << "found large size bit-vector:\n" << mk_pp(n, m) << "\n";);
-                    ctx.push_trail(value_trail<context, bool>(m_approximates_large_bvs));
+                    ctx.push_trail(value_trail<bool>(m_approximates_large_bvs));
                     m_approximates_large_bvs = true;
                 }
                 return true;
@@ -1100,16 +1101,16 @@ namespace smt {
     }
 
     void theory_bv::apply_sort_cnstr(enode * n, sort * s) {
-        if (!is_attached_to_var(n) && !approximate_term(n->get_owner())) {
+        if (!is_attached_to_var(n) && !approximate_term(n->get_expr())) {
             mk_bits(mk_var(n));
             if (ctx.is_relevant(n)) {
-                relevant_eh(n->get_owner());
+                relevant_eh(n->get_expr());
             }
         }
     }
     
     void theory_bv::new_eq_eh(theory_var v1, theory_var v2) {
-        TRACE("bv_eq", tout << "new_eq: " << mk_pp(get_enode(v1)->get_owner(), m) << " = " << mk_pp(get_enode(v2)->get_owner(), m) << "\n";);
+        TRACE("bv_eq", tout << "new_eq: " << mk_pp(get_enode(v1)->get_expr(), m) << " = " << mk_pp(get_enode(v2)->get_expr(), m) << "\n";);
         TRACE("bv", tout << "new_eq_eh v" << v1 << " = v" << v2 << " @ " << ctx.get_scope_level() << 
               " relevant1: " << ctx.is_relevant(get_enode(v1)) << 
               " relevant2: " << ctx.is_relevant(get_enode(v2)) << "\n";);
@@ -1175,7 +1176,7 @@ namespace smt {
 
         literal_vector & lits = m_tmp_literals;
         lits.reset();
-        literal eq = mk_eq(get_enode(v1)->get_owner(), get_enode(v2)->get_owner(), true);
+        literal eq = mk_eq(get_enode(v1)->get_expr(), get_enode(v2)->get_expr(), true);
         lits.push_back(eq);
         it1 = bits1.begin();
         it2 = bits2.begin();
@@ -1189,14 +1190,14 @@ namespace smt {
             lits.push_back(arg);
         }
         TRACE("bv", 
-              tout << mk_pp(get_enode(v1)->get_owner(), m) << " = " << mk_pp(get_enode(v2)->get_owner(), m) << " " 
+              tout << mk_pp(get_enode(v1)->get_expr(), m) << " = " << mk_pp(get_enode(v2)->get_expr(), m) << " " 
               << ctx.get_scope_level() 
               << "\n";
               ctx.display_literals_smt2(tout, lits););
 
         m_stats.m_num_diseq_dynamic++;
         scoped_trace_stream st(*this, lits);
-        ctx.mk_th_axiom(get_id(), lits.size(), lits.c_ptr());
+        ctx.mk_th_axiom(get_id(), lits.size(), lits.data());
     }
 
     void theory_bv::assign_eh(bool_var v, bool is_true) {
@@ -1312,7 +1313,7 @@ namespace smt {
                 ctx.mark_as_relevant(lits[2]);
                 {
                     scoped_trace_stream _sts(*this, lits);
-                    ctx.mk_th_axiom(get_id(), lits.size(), lits.c_ptr());
+                    ctx.mk_th_axiom(get_id(), lits.size(), lits.data());
                 }
             }
      
@@ -1341,6 +1342,7 @@ namespace smt {
     }
 
     void theory_bv::relevant_eh(app * n) {
+        TRACE("arith", tout << "relevant: #" << n->get_id() << " " << ctx.e_internalized(n) << ": " << mk_pp(n, m) << "\n";);
         TRACE("bv", tout << "relevant: #" << n->get_id() << " " << ctx.e_internalized(n) << ": " << mk_pp(n, m) << "\n";);
         if (m.is_bool(n)) {
             bool_var v = ctx.get_bool_var(n);
@@ -1441,7 +1443,7 @@ namespace smt {
         m_util(ctx.get_manager()),
         m_autil(ctx.get_manager()),
         m_bb(ctx.get_manager(), ctx.get_fparams()),
-        m_trail_stack(*this),
+        m_trail_stack(),
         m_find(*this),
         m_approximates_large_bvs(false) {
         memset(m_eq_activity, 0, sizeof(m_eq_activity));
@@ -1577,7 +1579,7 @@ namespace smt {
     void theory_bv::propagate() {
         if (!can_propagate())
             return;
-        ctx.push_trail(value_trail<context, unsigned>(m_prop_diseqs_qhead));
+        ctx.push_trail(value_trail<unsigned>(m_prop_diseqs_qhead));
         for (; m_prop_diseqs_qhead < m_prop_diseqs.size() && !ctx.inconsistent(); ++m_prop_diseqs_qhead) {
             auto p = m_prop_diseqs[m_prop_diseqs_qhead];
             assert_new_diseq_axiom(p.v1, p.v2, p.idx);
@@ -1621,7 +1623,7 @@ namespace smt {
             ast_manager & m = cr.get_manager();
             expr_ref fact(m);
             ctx.literal2expr(m_consequent, fact);
-            return m.mk_th_lemma(get_from_theory(), fact, prs.size(), prs.c_ptr());
+            return m.mk_th_lemma(get_from_theory(), fact, prs.size(), prs.data());
         }
 
         theory_id get_from_theory() const override {
