@@ -136,7 +136,7 @@ struct has_nlmul {
     has_nlmul(ast_manager& m):m(m), a(m) {}
     
     void throw_found(expr* e) {
-        TRACE("probe", tout << expr_ref(e, m) << ": " << sort_ref(m.get_sort(e), m) << "\n";);
+        TRACE("probe", tout << expr_ref(e, m) << ": " << sort_ref(e->get_sort(), m) << "\n";);
         throw found();
     }
 
@@ -356,7 +356,7 @@ static bool is_lp(goal const & g) {
         while (m.is_not(f, f))
             sign = !sign;
         if (m.is_eq(f) && !sign) {
-            if (m.get_sort(to_app(f)->get_arg(0))->get_family_id() != u.get_family_id())
+            if (to_app(f)->get_arg(0)->get_sort()->get_family_id() != u.get_family_id())
                 return false;
             continue;
         }
@@ -438,7 +438,7 @@ struct is_non_nira_functor {
     is_non_nira_functor(ast_manager & _m, bool _int, bool _real, bool _quant, bool linear):m(_m), u(m), m_int(_int), m_real(_real), m_quant(_quant), m_linear(linear) {}
 
     void throw_found(expr* e) {
-        TRACE("probe", tout << expr_ref(e, m) << ": " << sort_ref(m.get_sort(e), m) << "\n";);
+        TRACE("probe", tout << expr_ref(e, m) << ": " << sort_ref(e->get_sort(), m) << "\n";);
         throw found();
     }
 
